/*
 * Routines involved with enforcing the rules of Mille Bournes.
 */
#include "mille.h"


static	void	calcstat PROTO((PLAYER *, STATUS));


/*
 * See if a player can play one of his cards.
 * If the card can be legally played on this turn, NULL is returned.
 * Otherwise, the address of an appropriate error message is returned.
 */
char *
canplaymsg(pp, card)
	PLAYER	*pp;			/* player wanting to use card */
	CARD	card;			/* card being played */
{
	PLAYER		*op;		/* opposing player */
	int		cardindex;	/* index of card */
	static	char	buf[80];	/* message */

	op = pp->p_enemy;

	if (isnocard(card))		/* make sure have a card */
		return "No such card";

	cardindex = -1;

	do {
		if (++cardindex > HANDSIZE)
			return "Card not found";	/* card not in the hand */
	} while (pp->p_cards[cardindex] != card);

	if ((pp->p_cardcount <= HANDSIZE) && (cardsleft > 0)) {
		if (ishazard(pp->p_coupcard) && (card == hazardtosafety(pp->p_coupcard)))
			return NULL;

		return "Must pick first";
	}

	if (issafety(card))		/* can always play a safety card */
		return NULL;

	if (ismileage(card)) {		/* see if can use mileage card */
		if (pp->p_stopped) {
			if (isremedy(pp->p_battlepile) || isnocard(pp->p_battlepile) ||
				pp->p_safecards[safetyindex(hazardtosafety(pp->p_battlepile))])
					return "Need a Roll card";

			sprintf(buf, "Need to remedy %s", cardnames[pp->p_battlepile]);
			return buf;
		}

		if (pp->p_limited && (card > TC_50MILE))
			return "Speed Limit is 50";

		if ((pp->p_mileage + mileage[mileageindex(card)]) > mileagegoal) {
			sprintf(buf, "Exceeds goal of %d", mileagegoal);
			return buf;
		}

		if ((card == TC_200MILE) &&
			(pp->p_milecards[mileageindex(TC_200MILE)] >= 2))
				return "Can only use two 200 cards";

		return NULL;
	}

	if (card == TC_ROLL) {		/* see if we can use roll card */
		if (!pp->p_stopped)
			return "Not stopped";

		if (isnocard(pp->p_battlepile))
			return NULL;

		if ((pp->p_battlepile == TC_STOP) || isremedy(pp->p_battlepile))
			return NULL;

		if (pp->p_safecards[safetyindex(hazardtosafety(pp->p_battlepile))])
			return NULL;

		sprintf(buf, "Need to remedy %s", cardnames[pp->p_battlepile]);
		return buf;
	}

	if (card == TC_ENDOFLIMIT) {	/* see if can end speed limit */
		if (pp->p_limited)
			return NULL;

		return "No Speed Limit in effect";
	}

	if (isremedy(card)) {		/* see if can remedy current hazard */
		if (!pp->p_stopped)
			return "Not stopped";

		if (!ishazard(pp->p_battlepile))
			return "No hazard to remedy";

		if (pp->p_battlepile == remedytohazard(card))
			return NULL;

		sprintf(buf, "Wrong remedy for %s", cardnames[pp->p_battlepile]);
		return buf;
	}

	if (card == TC_SPEEDLIMIT) {	/* see if can play speed limit */
		if (op->p_safecards[safetyindex(TC_RIGHTOFWAY)])
			return "Protected by Right of Way";

		return NULL;
	}

	if (ishazard(card)) {		/* see if can play hazard */
		if (op->p_stopped && isnocard(op->p_battlepile))
			return "Opponent must play Roll";

		if (!op->p_safecards[safetyindex(hazardtosafety(card))])
			return NULL;

		sprintf(buf, "Protected by %s", cardnames[hazardtosafety(card)]);
		return buf;
	}

	return "Unknown card";
}


/*
 * Play the specified card.
 * It is assumed that canplay was previously called to verify that the
 * use of the card is legal, and so legality checks are not made here.
 */
void
playcard(pp, card)
	PLAYER	*pp;			/* player using card */
	CARD	card;			/* card being played */
{
	PLAYER	*op;			/* opponent */
	CARD	coupcard;		/* coup fourre card */

	coupcard = pp->p_coupcard;
	discard(pp, card);

	if (ismileage(card)) {		/* play mileage card */
		pp->p_mileage += mileage[mileageindex(card)];
		pp->p_milecards[mileageindex(card)]++;
		score(pp);
		return;
	}

	if (card == TC_ENDOFLIMIT) {	/* end of limit */
		pp->p_speedpile = card;
		pp->p_limited = FALSE;
		return;
	}

	if (isremedy(card)) {		/* remedy a hazard */
		pp->p_battlepile = card;
		checkstop(pp);
		return;
	}

	if (issafety(card)) {		/* play a safety */
		pp->p_moved = FALSE;	/* this is a free turn */
		pp->p_safecards[safetyindex(card)] = TRUE;
		checkstop(pp);
		score(pp);

		/*
		 * Check for coup fourre
		 */
		if (isnocard(coupcard) || (card != hazardtosafety(coupcard)))
			return;

		pp->p_coupcards[safetyindex(card)] = TRUE;
		score(pp);

		if (coupcard == TC_SPEEDLIMIT) {
			pp->p_speedpile = pp->p_oldspeed;
			return;
		}

		pp->p_battlepile = pp->p_oldbattle;
		checkstop(pp);
		return;
	}

	op = &players[1 - pp->p_index];

	if (card == TC_SPEEDLIMIT) {	/* play a speed limit on the opponent */
		op->p_oldspeed = op->p_speedpile;
		op->p_speedpile = TC_SPEEDLIMIT;
		op->p_coupcard = TC_SPEEDLIMIT;
		op->p_limited = TRUE;
		return;
	}

	if (ishazard(card)) {		/* play a hazard on the opponent */
		op->p_oldbattle = op->p_battlepile;
		op->p_battlepile = card;
		op->p_coupcard = card;
		op->p_stopped = TRUE;
		return;
	}
}


/*
 * Called when something happens which can affect start/stop of player.
 * Sets p_stopped to indicate whether or not the player is stopped, and
 * p_limited to indicate whether or not speed limit is in effect.
 */
void
checkstop(pp)
	PLAYER	*pp;
{
	CARD	card;
	int	rightofway;

	rightofway = pp->p_safecards[safetyindex(TC_RIGHTOFWAY)];
	pp->p_limited = ((pp->p_speedpile == TC_SPEEDLIMIT) && !rightofway);
	card = pp->p_battlepile;

	if (card == TC_ROLL) {
		pp->p_stopped = FALSE;
		return;
	}

	if (ishazard(card) && !pp->p_safecards[safetyindex(hazardtosafety(card))]) {
		pp->p_stopped = TRUE;
		return;
	}

	pp->p_stopped = !rightofway;
}


/*
 * Determine the card needed to allow a player to play good mileage.
 * This is either the proper remedy card if the player is stopped,
 * or an end of limit card if a speed limit is in effect.
 */
CARD
fixcard(pp)
	PLAYER	*pp;		/* player to check */
{
	CARD	card;		/* card to be fixed */

	if (!pp->p_stopped) {
		if (pp->p_limited)
			return TC_ENDOFLIMIT;	/* need end of limit */
		return TC_NONE;		/* not stopped */
	}

	card = pp->p_battlepile;

	if (!ishazard(card) || (pp->p_safecards[safetyindex(hazardtosafety(card))]))
		return TC_ROLL;		/* need to get a roll */

	return hazardtoremedy(card);	/* need this remedy */
}


/*
 * Remove the specified card from the player's hand.
 * This simply turns it into a nonexistent card, and resorts the hand.
 */
void
discard(pp, card)
	PLAYER	*pp;		/* player whose card is discarded */
	CARD	card;
{
	int	cardindex;	/* card to remove */

	pp->p_moved = TRUE;
	pp->p_coupcard = TC_NONE;
	pp->p_pickindex = -1;

	if (isnocard(card))
		return;			/* not a real card */

	cardindex = -1;
	do {
		if (++cardindex > HANDSIZE)
			return;		/* card not in the hand */
	} while (pp->p_cards[cardindex] != card);

	pp->p_cards[cardindex] = TC_NONE;
	pp->p_cardcount--;
	typesleft[card]--;

	sorthand(pp);
}


/*
 * Check to see if the hand is finished, and return status bits indicating
 * the result of the hand or game, from the point of view of the player.
 * Status is zero if the hand is not yet over.	The hand is over when
 * one player reaches the mileage goal and doesn't extend, or when we can
 * determine that no more points can be played for both players.
 */
STATUS
gamestatus()
{
	PLAYER	*pp;		/* current player examined */
	STATUS	status;		/* status bits to return */
	STATUS	machstat;
	CARD	card;
	CARD	lowcard;
	BOOL	gotroll;
	BOOL	gotendlimit;
	BOOL	gotremedy;
	int	score1;
	int	score2;
	int	i;

	status = 0;
	for (pp = players; pp < &players[2]; pp++) {
		if (pp->p_mileage < mileagegoal)
			continue;

		if ((mileagegoal == S_SHORTGOAL) && ((*pp->p_extend)(pp))) {
			mileagegoal = S_LONGGOAL;
			return 0;
		}

		pp->p_tripdone = TRUE;
		score(pp);
		status |= ST_TRIPDONE;
		goto handover;
	}

	if (cardsleft > 0)
		return 0;		/* still have cards in deck */

	for (pp = players; pp < &players[2]; pp++) {
		gotroll = FALSE;
		gotendlimit = FALSE;
		gotremedy = FALSE;
		lowcard = 1000;

		for (i = 0; i < pp->p_cardcount; i++) {
			card = pp->p_cards[i];

			if (issafety(card))
				return 0;	/* safeties can always be used */

			if (card == TC_ROLL)
				gotroll = TRUE;

			if (card == TC_ENDOFLIMIT)
				gotendlimit = TRUE;

			if (ishazard(pp->p_battlepile) &&
				(card == hazardtoremedy(pp->p_battlepile)))
					gotremedy = TRUE;

			if (ismileage(card) && (card < lowcard))
				lowcard = card;
		}

		if (!ismileage(lowcard))
			continue;	/* no mileage cards */

		if ((pp->p_mileage + mileage[mileageindex(lowcard)]) > mileagegoal)
			continue;	/* mileages too high for goal */

		if (pp->p_limited && !gotendlimit && (lowcard > TC_50MILE))
			continue;	/* mileages too high for speed limit */

		if (!pp->p_stopped)
			return 0;	/* can play a mileage card */

		if (!gotroll && !pp->p_safecards[safetyindex(TC_RIGHTOFWAY)])
			continue;	/* stopped and no roll or right of way */

		card = pp->p_battlepile;
		if (gotremedy || isnocard(card) || isremedy(card))
			return 0;	/* can remedy, roll, and do mileage */

		if (pp->p_safecards[safetyindex(hazardtosafety(card))])
			return 0;	/* can use roll and do mileage */
	}

	/*
	 * No progress can be made.  End the hand.
	 * Here when the hand is over.	See if the game is over too.
	 * Add in statistics for our convenience.
	 */
handover:
	status |= ST_HANDOVER;
	if (player->p_score > machine->p_score)
		status |= ST_HANDWON;

	if (player->p_score == machine->p_score)
		status |= ST_HANDTIED;

	savedata.sd_hands++;
	score1 = player->p_gamescore;
	score2 = machine->p_gamescore;

	if (((score1>=S_GAMEGOAL) || (score2>=S_GAMEGOAL)) && (score1 != score2)) {
		status |= ST_GAMEOVER;

		if (score1 > score2)
			status |= ST_GAMEWON;

		savedata.sd_games++;
	}

	calcstat(player, status);

	machstat = (status & ST_GAMEOVER);

	if ((status & ST_GAMEOVER) && ((status & ST_GAMEWON) == 0))
		machstat |= ST_GAMEWON;

	if ((status & (ST_HANDWON | ST_HANDTIED)) == 0)
		machstat |= ST_HANDWON;

	calcstat(machine, machstat);

	return status;
}


/*
 * Calculate statistics for a player.
 */
static void
calcstat(pp, status)
	PLAYER	*pp;		/* player who gets statistics */
	STATUS	status;		/* status from my point of view */
{
	int	i;		/* index */
	int	score;		/* our game score */

	if (pp->p_besthand < pp->p_score)
		pp->p_besthand = pp->p_score;

	if (status & ST_HANDWON) {
		pp->p_wonhands++;

		if ((status & ST_TRIPDONE) && (pp->p_enemy->p_mileage == 0))
			pp->p_shutouts++;
	}

	if (status & ST_GAMEOVER) {
		score = pp->p_gamescore;

		for (i = 8; i >= 0; i--)
			pp->p_tengamescores[i + 1] = pp->p_tengamescores[i];

		pp->p_tengamescores[0] = score;
		pp->p_allgamescore += score;

		if ((pp->p_worstgame > score) || (pp->p_bestgame == 0))
			pp->p_worstgame = pp->p_gamescore;

		if (pp->p_bestgame < score)
			pp->p_bestgame = score;

		if (status & ST_GAMEWON)
			pp->p_games++;
	}
}


/*
 * Calculate the current score of a player.
 * The result is saved in the player structure.
 */
void
score(pp)
	PLAYER	*pp;		/* player to calculate */
{
	int	score;
	int	i;

	score = 0;

	for (i = 0; i < SAFETYPES; i++)	/* per safety */
		if (pp->p_safecards[i]) score += S_SAFETY;

	if (score == (SAFETYPES * S_SAFETY))	/* all four safeties */
		score += S_ALLSAFETY;

	for (i = 0; i < SAFETYPES; i++)	/* per coup fourre */
		if (pp->p_coupcards[i]) score += S_COUPFOURRE;

	score += pp->p_mileage;		/* mileage played */

	if (!pp->p_tripdone) {		/* done if not finished trip */
		if (score > pp->p_score)
			pp->p_gamescore += (score - pp->p_score);

		pp->p_score = score;
		return;
	}

	score += S_TRIPDONE;		/* trip complete */

	if (cardsleft == 0)		/* delayed action */
		score += S_DELAYED;

	if (pp->p_milecards[mileageindex(TC_200MILE)] == 0)	/* safe trip */
		score += S_SAFETRIP;

	if (mileagegoal == S_LONGGOAL)	/* extension */
		score += S_EXTENSION;

	if (pp->p_enemy->p_mileage == 0)	/* shut out */
		score += S_SHUTOUT;

	if (score > pp->p_score)
		pp->p_gamescore += (score - pp->p_score);

	pp->p_score = score;
}


/*
 * Try to add the specified number of cards to the player's hand.
 * The cards are added to the end of the hand, and then resorted.
 * Returns error string if no picking is impossible, or NULL if at
 * least one card can be drawn.
 */
char *
pickcards(pp, count)
	PLAYER	*pp;		/* player to pick card for */
	int	count;
{
	int	num;		/* number of cards actually got */

	pp->p_coupcard = TC_NONE;
	for (num = 0; num < count; num++) {
		if (cardsleft <= 0) {
			if (num)
				break;

			return "No cards left in deck";
		}

		if ((pp->p_cardcount > HANDSIZE) || (iscard(pp->p_cards[HANDSIZE]))) {
			if (num)
				break;

			return "Your hand is full";
		}

		pp->p_cards[HANDSIZE] = carddeck[--cardsleft];
		pp->p_cardcount++;
		pp->p_pickindex = HANDSIZE;

		sorthand(pp);
	}

	return NULL;
}


/*
 * Sort a player's hand so that cards of the same type appear together.
 * The card just picked is also sorted.	The rankings are such that
 * blank cards are placed at the end.
 */
void
sorthand(pp)
	PLAYER	*pp;		/* player whose hand is sorted */
{
	int	i;
	int	j;
	CARD	*t1;
	CARD	*t2;
	CARD	tmp;

	for (i = HANDSIZE - 1; i >= 0; i--) {
		for (j = 0; j <= i; j++) {
			t1 = &pp->p_cards[j];
			t2 = &pp->p_cards[j+1];

			if (cardranks[*t1] > cardranks[*t2]) {
				tmp = *t1;
				*t1 = *t2;
				*t2 = tmp;

				if ((pp->p_pickindex == j) || (pp->p_pickindex == (j+1)))
					pp->p_pickindex = (2 * j + 1) - pp->p_pickindex;
			}
		}
	}
}

/* END CODE */
